/**
 * Copyright (C) 2008-2012 AgroSense Foundation.
 *
 * AgroSense is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * There are special exceptions to the terms and conditions of the GPLv3 as it is applied to
 * this software, see the FLOSS License Exception
 * <http://www.agrosense.eu/foss-exception.html>.
 *
 * AgroSense is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with AgroSense.  If not, see <http://www.gnu.org/licenses/>.
 */
package nl.cloudfarming.client.field.model;

import com.vividsolutions.jts.geom.Envelope;
import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.geom.Point;
import java.awt.Image;
import java.awt.datatransfer.DataFlavor;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.IOException;
import java.io.Serializable;
import java.util.*;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.swing.Action;
import nl.cloudfarming.client.geoviewer.DynamicGeometrical;
import nl.cloudfarming.client.lib.geotools.GeometryTools;
import nl.cloudfarming.client.model.AbstractEntity;
import nl.cloudfarming.client.model.EnterpriseIdProvider;
import nl.cloudfarming.client.model.ItemIdType;
import nl.cloudfarming.client.model.Temporal;

/**
 * Class which represents fields Listens to propertyChanges on it's
 * ActivityFields if there are changes will fire a PropertyChangeEvent
 *
 * @author Timon Veenstra <monezz@gmail.com>
 */
public class Field extends AbstractEntity implements DynamicGeometrical<Geometry>, Serializable, PropertyChangeListener, Temporal {

    private static final Logger LOGGER = Logger.getLogger("nl.cloudfarming.client.field.model.Field");
    private static final long serialVersionUID = 8763622679187376702L;
    public static final ItemIdType ITEM_ID_TYPE = ItemIdType.FLD;
    public static final String EXT = "fld";
    public static final String MIMETYPE = "application/x-agrosense-field";
    public static final DataFlavor DATA_FLAVOR = new DataFlavor(Field.class, "Field");
    private Geometry geometry;
    public static final String PROP_GEOMETRY = "geometry";
    private String name;
    public static final String PROP_NAME = "name";
    private double areaInHectares = 0.0d;
    public static final String PROP_AREA_IN_HECTARES = "areaInHectares";
    private Date startDate;
    public static final String PROP_START_DATE = "startDate";
    private Date endDate;
    public static final String PROP_END_DATE = "endDate";
    private SoilType soilType;
    public static final String PROP_SOIL_TYPE = "soilType";
    private transient List<Action> mapActions = null;
    protected Set<ActivityField> activityFields = new HashSet<>();
    public static final String PROP_ACTIVITYFIELDS = "ActivityFields";

    /**
     * default field constructor
     */
    public Field(EnterpriseIdProvider eip) {
        super(eip, ITEM_ID_TYPE);
        if (this.geometry != null) {
            this.areaInHectares = GeometryTools.getAreaInHA(geometry);
        }
    }

    /**
     * name constructor
     */
    public Field(EnterpriseIdProvider eip, String name) {
        this(eip);
        this.name = name;
    }
    
    
    /**
     * copy constructor.
     * @param other 
     */
    public Field(Field other) {
        this(other.getURI());
        this.name = other.getName();
        this.geometry = other.getGeometry();
        this.startDate = other.getStartDate();
        this.endDate = other.getEndDate();
        this.soilType = other.getSoilType();
    }

    /**
     * Get the value of ActivityFields
     *
     * @return unmodifiable version of ActivityFields list
     */
    public Set<ActivityField> getActivityFields() {
        if (activityFields == null) {
            activityFields = new HashSet<>();
        }
        return Collections.unmodifiableSet(activityFields);
    }

    /**
     * Get an activity field by its name will return null when no activity field
     * with given name is found
     *
     * @param name
     * @return activity field or null when nothing found
     */
    public ActivityField getActivityFieldByName(String name) {
        for (ActivityField activityField : getActivityFields()) {
            if (name.equalsIgnoreCase(activityField.getName())) {
                return activityField;
            }
        }
        return null;
    }

    /**
     * Add a ActivityField to the groups in this field
     *
     * @param activityField
     */
    public void addActivityField(ActivityField activityField) {
        if (activityFields == null) {
            activityFields = new HashSet<>();
        }
        Set<ActivityField> oldActivityFields = new HashSet<>(this.activityFields);
        LOGGER.log(Level.FINEST, "addActivityField, pre add activityFields containts {0} entries", activityFields.size());
        activityFields.add(activityField);
        activityField.addPropertyChangeListener(this);
        LOGGER.log(Level.FINEST, "addActivityField, post add activityFields containts {0} entries", activityFields.size());
        getPropertyChangeSupport().firePropertyChange(PROP_ACTIVITYFIELDS, oldActivityFields, new HashSet<>(this.activityFields));
    }

    /**
     * Remove a ActivityField from the groups in this field.
     *
     * @param activityField
     */
    public void removeActivityField(ActivityField activityField) {
        if (activityFields != null && activityFields.contains(activityField)) {
            Set<ActivityField> oldActivityFields = new HashSet<>(this.activityFields);
            activityFields.add(activityField);
            activityField.removePropertyChangeListener(this);
            getPropertyChangeSupport().firePropertyChange(PROP_ACTIVITYFIELDS, oldActivityFields, new HashSet<>(this.activityFields));
        }
    }

    /**
     * Get the value of endDate
     *
     * @return the value of endDate
     */
    public Date getEndDate() {
        return endDate;
    }

    /**
     * Set the value of endDate
     *
     * @param endDate new value of endDate
     */
    public void setEndDate(Date endDate) {
        Date oldEndDate = this.endDate;
        this.endDate = endDate;
        getPropertyChangeSupport().firePropertyChange(PROP_END_DATE, oldEndDate, endDate);
    }

    /**
     * Get the value of startDate
     *
     * @return the value of startDate
     */
    public Date getStartDate() {
        return startDate;
    }

    /**
     * Set the value of startDate
     *
     * @param startDate new value of startDate
     */
    public void setStartDate(Date startDate) {
        Date oldStartDate = this.startDate;
        this.startDate = startDate;
        getPropertyChangeSupport().firePropertyChange(PROP_START_DATE, oldStartDate, startDate);
    }

    /**
     * Get the value of areaInHectares
     *
     * @return the value of areaInHectares
     */
    public double getAreaInHectares() {
        return areaInHectares;
    }

    /**
     * Set the value of areaInHectares
     *
     * @param areaInHectares new value of areaInHectares
     */
    public final void setAreaInHectares(double areaInHectares) {
        double oldAreaInHectares = this.areaInHectares;
        this.areaInHectares = areaInHectares;
        getPropertyChangeSupport().firePropertyChange(PROP_AREA_IN_HECTARES, oldAreaInHectares, areaInHectares);
    }

    /**
     * Get the soil type of this field
     *
     * @return
     */
    public SoilType getSoilType() {
        return soilType;
    }

    /**
     * Set the soil type of this field
     */
    public void setSoilType(SoilType soilType) {
        SoilType oldSoilType = this.soilType;
        this.soilType = soilType;
        getPropertyChangeSupport().firePropertyChange(PROP_SOIL_TYPE, oldSoilType, soilType);
    }

    /**
     * Get the value of geometry
     *
     * @return the value of geometry
     */
    @Override
    public Geometry getGeometry() {
        return geometry;
    }

    /**
     * Set the value of geometry
     *
     * @param geometry new value of geometry
     */
    @Override
    public void setGeometry(Geometry geometry) {
        Geometry oldGeometry = this.geometry;
        this.geometry = geometry;
        if (geometry != null) {
            setAreaInHectares(GeometryTools.getAreaInHA(geometry));
        }
        getPropertyChangeSupport().firePropertyChange(PROP_GEOMETRY, oldGeometry, geometry);
    }

    /**
     * Get the value of name
     *
     * @return the value of name
     */
    public String getName() {
        return name;
    }

    /**
     * Set the value of name
     *
     * @param name new value of name
     */
    public void setName(String name) {
        String oldName = this.name;
        this.name = name;
        getPropertyChangeSupport().firePropertyChange(PROP_NAME, oldName, name);
    }

    @Override
    public Point getCentroid() {
        return geometry == null ? null : geometry.getCentroid();
    }

    @Override
    public Geometry getRenderObject(Envelope envelope) {
        return geometry;
    }

    @Override
    public String getTooltipText() {
        return getName();
    }

    @Override
    public String toString() {
        return this.name;
    }

    @Override
    public Geometry getBoundingBox() {
        return (this.geometry != null) ? geometry.getEnvelope() : null;
    }

    @Override
    public Image getIcon() {
        return null;
    }

    @Override
    public String getIconLabel() {
        return null;
    }

    @Override
    public Action[] getMapActions() {
        return getActions().toArray(new Action[0]);
    }

    private List<Action> getActions() {
        if (mapActions == null) {
            mapActions = new ArrayList<>();
//            mapActions.add(CreateGeometryAction.get(this));
//            mapActions.add(EditGeometryAction.get(this));
        }
        return mapActions;
    }

    @Override
    public void addMapAction(Action action) {
        getActions().add(action);
    }

    private void readObject(java.io.ObjectInputStream in)
            throws IOException, ClassNotFoundException {
        in.defaultReadObject();
        for (ActivityField activityField : activityFields) {
            activityField.addPropertyChangeListener(this);
        }
    }

    /**
     *
     * @param pce
     */
    @Override
    public void propertyChange(PropertyChangeEvent pce) {
        getPropertyChangeSupport().firePropertyChange(PROP_ACTIVITYFIELDS, null, pce.getSource());
    }
    
    @Override
    public List<Temporal> getTemporalChildren() {
        //TODO here should be returned all task planned over field
        return new ArrayList<Temporal>(); 
    }
}
